# Input: some tex code, number of prior lines of tex (to allow absolute tex line
#        addressing for deeper code), the current indent level
# Value: if fully processed into shiny code, a list of length 1 with elements
#           'uiCode' and 'serverCode'
#        otherwise a list of the same form as the doList, with only the new elements
# Details: called by downOne() to further parse a given outer environment into a "doList"
#
processTexRange = function(tex, before, indent, eNum, globalInfoEnv) {
  lineNumOffset = get("lineNumOffset", envir=globalInfoEnv)
  # convert oridinary enumerate/itemize to shiny equivalents
  tex = gsub(paste0('[[:blank:]]*[\\](begin|end)[[:blank:]]*[{]enumerate[}]'), '\\\\\\1{shinyEnumerate}', tex)
  tex = gsub(paste0('[[:blank:]]*[\\](begin|end)[[:blank:]]*[{]itemize[}]'), '\\\\\\1{shinyItemize}', tex)
  
  # latex content environments to process
  outerEnvirons = paste0("(",
                         paste("shinySidePanel", "shinyMainPanel", "shinyColumns", 
                               "shinyNavbarMenu", "shinyConditionalPanel", 
                               "shinyEnumerate", "shinyItemize", sep="|"),
                         ")")
  BOE.RE = paste0("[\\]begin[[:blank:]]*[{]", outerEnvirons, "+[}]")
  
  # find outer content (tex) environments
  texLength = length(tex)
  outer = NULL # will be a data.frame with (start, end, environment name, option)
  offset = 0
  BOEs = grep(BOE.RE, tex)
  while (length(BOEs)>0) {
    # find name of first \begin{shinyFoo} environment
    start = BOEs[1] + offset
    temp = regexpr(BOE.RE, tex[start])
    environ = regmatches(tex[start], temp)
    temp2 = regexpr("shiny[[:alpha:]]+[}]", environ)
    environ = substring(environ, temp2, temp2+attr(temp2,"match.length")-2)
    # now get optional {} info
    tail = substring(tex[start], temp+attr(temp,"match.length")-1)
    tail = regmatches(tail, regexpr("[{].*?[}]", tail))
    if (length(tail)==0) option = NA_character_ else option = substring(tail, 2, nchar(tail)-1)
    # and now find ending tex code
    EOE.RE = paste0("[\\]end[[:blank:]]*[{]", environ, "[}]")
    downText = tex[start:texLength]
    assure(length(downText)>1, 283771)
    EOEs = grep(EOE.RE, downText[-1])
    if (length(EOEs) == 0) stop('no \\end{', environ, '}')
    BoThisEnvType = grep(paste0("[\\]begin[[:blank:]]*[{]",environ,"[}]"), downText)
    Seq = rep(0, length(downText))
    Seq[BoThisEnvType] = 1
    Seq[EOEs+1] = -1
    cumSeq = cumsum(Seq)
    if (!any(cumSeq==0)) {
      stop("@", lineNumOffset+start-1, "-", lineNumOffset+start+length(downText),
          " missing \\end{", environ, "}")
    }
    end = start -1 + min(which(cumSeq==0))
    outer = rbind(outer, data.frame(start=start, end=end, environ=environ, option=option))
    if (end == texLength) break
    offset = end
    BOEs = grep(BOE.RE, tex[(end+1):texLength])
  }
  
  # First check for all "open" code  -- later: pass through R comments only --
  if (is.null(outer)) return(list(innerCode(tex, eNum, before, globalInfoEnv)))
  
  ### Special: expand begin/end{shinyColumns} blocks to handle \shinyNextColumn
  colLines = which(outer$environ == "shinyColumns")
  if (length(colLines)>0) {
    for (r in rev(colLines)) { # end each shinyColumns block row (working bottom up)
      tempR = outer[r,]
      tempT = tex[tempR$start:tempR$end]
      tempN = grep("[[:blank:]]*[\\]shinyNextColumn", tempT)
      nAdd = length(tempN)
      if (nAdd>0) {
        temp = regexpr("shinyNextColumn[[:blank:]]*[{].+?[}]", tempT[tempN])
        if (any(temp==-1)) {
          stop("@", tempN[temp==-1][1]+tempR[1,"start"]-1+before+lineNumOffset,
               " \\shinyNextColumn requires a width in braces")
        }
        option = regmatches(tempT[tempN], temp)
        option = regmatches(option, regexpr("[{].+[}]", option))
        option = substring(option, 2, nchar(option)-1)
        noption = suppressWarnings(as.numeric(option))
        badOption = suppressWarnings(is.na(noption) | noption!=round(noption) | noption<1 | noption>12)
        if (any(badOption)) {
          rng = grep("[\\](begin|end)[[:blank:]]*[{]shinyColumns[}]", tempT)+tempR[1,"start"]-1+before+lineNumOffset
          stop("@", rng[1], "-", rng[2],
               " \\shinyNextColumn requires a width that is a whole number from 1 to 12")
        }
        tempD = data.frame(start=tempR$start+tempN-1, end=NA, environ=I("shinyNextColumn"), 
                           option=option)
        tempD[nAdd, "environ"] = "shinyLastColumn"
        if (r==nrow(outer)) {
          outer = rbind(outer[1:r,], tempD)
        } else {
          outer = rbind(outer[1:r,], tempD, outer[(r+1):nrow(outer),])          
        }
        outer[r+(0:nAdd),"end"] = c(outer[r+(1:nAdd),"start"], outer[r,"end"])
      } # end adding shinyNextColumn sections
    } # end for each shinyColumns block row (working bottom up)
    allCols = c("shinyColumns","shinyNextColumn","shinyLastColumn")
    colSum = sum(as.numeric(as.character(outer[outer$environ %in% allCols, "option"])))
    if (colSum > 12) {
      rng  = c(outer[outer$environ=="shinyColumns", "start"], 
               outer[outer$environ=="shinyLastColumn", "end"])
      assure(length(rng)==2, 878261)
      rng = rng + before + lineNumOffset
      stop("@", rng[1], "-", rng[2], " \\shinyColumn widths add to more than 12")
    }
  }
  
  ### Now complete 'outer' to include code outside of 'shinyFoo' environments
  # Then fill "open" code segments into 'outer'
  comments = union(grep("^[[:blank:]]*$", tex), grep("^[[:blank:]]*%", tex))
  # Possibly add a line at the top of 'outer' for an initial segment
  if (outer[1,"start"]>1) {
    top = 1:(outer[1,"start"]-1)
    if (length(setdiff(top, comments))>0) {
      outer = rbind(data.frame(start=1, end=outer[1,"start"]-1, environ=NA, option=NA), outer)
    }
  }
  #
  # Then setup in-between segments
  nOuter = nrow(outer)  # nrow(outer) changes in the loop!
  outer$orig = 1:nOuter
  if (nOuter>=2) {
    for (i in 2:nOuter) {
      row = which(outer$orig==i)
      # next line cannot be "==" due to special handling of //shinyNextColumn
      if (outer[row-1, "end"] + 1 >= outer[row, "start"]) next
      range = (outer[row-1, "end"] + 1) : (outer[row, "start"] - 1)
      if (length(setdiff(range, comments))>0) {
        outer = rbind(outer[1:(row-1),], 
                      data.frame(start=range[1], end=range[length(range)], environ=NA, option=NA, orig=NA),
                      outer[row:nrow(outer),])
      }
    } # end for each "between"
  } # end if any betweens
  #
  # And finally setup the terminal segment, if any
  end = outer[nrow(outer), "end"]
  if (end < texLength) {
    range = (end + 1) : texLength
    if (length(setdiff(range, comments))>0) {
      outer = rbind(outer, 
                    data.frame(start=end+1, end=texLength, environ=NA, option=NA, orig=NA))
    }
  }

  ### Now construct a doList as in startDoList()
  lst = NULL
  for (i in 1:nrow(outer)) {
    if (is.na(outer[i, "environ"])) {
      temp = list(uiCode=as.numeric(c(before+outer[i,"start"], before+outer[i,"end"])), 
                  serverCode=NA, nextElement=NA, indent=indent)
      lst = c(lst, list(temp))
    }  else {
      offset = 1 - length(grep("shiny(Enumerate|Itemize)", outer[i,"environ"]) == 1)
      temp = list(list(uiCode = prefaceCode(outer[i,"environ"], outer[i,"option"],
                                            lineNumOffset=lineNumOffset+before+outer[i,"start"],
                                            globalInfoEnv=globalInfoEnv),
                       serverCode = NA, nextElement=NA, indent=indent),
                  list(uiCode = before + c(outer[i,"start"]+1, outer[i,"end"]-1),
                       serverCode = NA, nextElement=NA, indent=indent+offset),
                  list(uiCode = suffixCode(outer[i,"environ"], outer[i,"option"]),
                       serverCode = NA, nextElement=NA, indent=indent))
      lst = c(lst, temp)
    }
  }
  return(lst)
} #end processTexRange()
